package Language::INTERCAL::ByteCode;

# Definitions of bytecode symbols etc

# This file is part of CLC-INTERCAL

# Copyright (c) 2006-2008, 2023 Claudio Calvelli, all rights reserved.

# CLC-INTERCAL is copyrighted software. However, permission to use, modify,
# and distribute it is granted provided that the conditions set out in the
# licence agreement are met. See files README and COPYING in the distribution.


use strict;
use vars qw($VERSION $PERVERSION $DATAVERSION);
($VERSION) = ($PERVERSION = "CLC-INTERCAL/Base INTERCAL/ByteCode.pm 1.-94.-2.4") =~ /\s(\S+)$/;

use Carp;
use Language::INTERCAL::Exporter '1.-94.-3', qw(import compare_version);
use Language::INTERCAL::Splats '1.-94.-2.1',
	qw(faint SP_INTERNAL SP_BCMATCH SP_TODO);
use Language::INTERCAL::RegTypes '1.-94.-2.2',
    qw(REG_spot REG_twospot REG_tail REG_hybrid REG_dos REG_whp REG_shf REG_cho);

$DATAVERSION = '1.-94.-2.5';
compare_version($VERSION, $DATAVERSION) < 0 and $VERSION = $DATAVERSION;

use constant BYTE_SIZE     => 8;      # number of bits per byte (must be == 8)
use constant NUM_OPCODES   => 0x80;   # number of virtual opcodes
use constant OPCODE_RANGE  => 1 << BYTE_SIZE;
use constant BYTE_SHIFT    => OPCODE_RANGE - NUM_OPCODES;

use vars qw(@EXPORT_OK %EXPORT_TAGS);
@EXPORT_OK = qw(
    bytecode bytedecode bytename bc_list BC BCget is_constant
    bc_skip bc_forall add_bytecode all_opcodes NUM_OPCODES
    BC_ABG BC_ABL BC_AND BC_AWC BC_BIT BC_BLM BC_BUG BC_BUT BC_BWC BC_CFG
    BC_CFL BC_CHO BC_CON BC_CRE BC_CWB BC_DES BC_DIV BC_DOS BC_DSX BC_EBC
    BC_ECB BC_ENR BC_FIN BC_FLA BC_FOR BC_FRZ BC_GRA BC_GUP BC_HSN BC_HYB
    BC_IGN BC_INT BC_IOR BC_LAB BC_LEA BC_MKB BC_MSP BC_MUL BC_NLB BC_NOT
    BC_NUM BC_NXG BC_NXL BC_NXT BC_OSN BC_OVM BC_OVR BC_QUA BC_REG BC_REL
    BC_REM BC_REO BC_RES BC_RET BC_RIN BC_ROU BC_RSE BC_SEL BC_SHF BC_SPL
    BC_SPO BC_STA BC_STO BC_STR BC_STS BC_STU BC_SUB BC_SWA BC_SWB BC_SYS
    BC_TAI BC_TRD BC_TRU BC_TSP BC_TYP BC_UDV BC_UNA BC_UNE BC_UNS BC_WHP
    BC_WIN
);

%EXPORT_TAGS = (
    BC => [qw(
	BC_ABG BC_ABL BC_AND BC_AWC BC_BIT BC_BLM BC_BUG BC_BUT BC_BWC
	BC_CFG BC_CFL BC_CHO BC_CON BC_CRE BC_CWB BC_DES BC_DIV BC_DOS
	BC_DSX BC_EBC BC_ECB BC_ENR BC_FIN BC_FLA BC_FOR BC_FRZ BC_GRA
	BC_GUP BC_HSN BC_HYB BC_IGN BC_INT BC_IOR BC_LAB BC_LEA BC_MKB
	BC_MSP BC_MUL BC_NLB BC_NOT BC_NUM BC_NXG BC_NXL BC_NXT BC_OSN
	BC_OVM BC_OVR BC_QUA BC_REG BC_REL BC_REM BC_REO BC_RES BC_RET
	BC_RIN BC_ROU BC_RSE BC_SEL BC_SHF BC_SPL BC_SPO BC_STA BC_STO
	BC_STR BC_STS BC_STU BC_SUB BC_SWA BC_SWB BC_SYS BC_TAI BC_TRD
	BC_TRU BC_TSP BC_TYP BC_UDV BC_UNA BC_UNE BC_UNS BC_WHP BC_WIN
    )],
);

my @bytecodes = (
    ['STart of Statement', 'S', 'STS', '###C(#)S', 0, 0, "Takes a variable number of constants, not less than four. The first constant indicates the byte position in the source code where this statement was compiled from; the second constant indicates the length of the statement in the source code; the third indicates whether the statement may be a comment (it has not been recognised using the currently active grammar) or not; the fourth indicates the number of constants following. The rest of the constants indicate which grammar rules were used to compile this particular statement.\n\nAt runtime, not all grammar rules may be available at all times, depending on the history of I<CRE> and I<DES>. To execute a statement corresponding to a given bit of source code, the runtime will find all relevant I<STS> statements find the best one which could have been compiled given the current state of the grammar, and executes it; if a non-comment statement is available, it will be used, otherwise a comment one will have to do. If execution is to proceed sequentially, the second constant is used to figure out how to repeat this process. Execution starts at byte offset 0 in the source code.\n\nGrammar rules are numbered at compile time, and may differ from program to program."], # 0
    ['STOre', 'S', 'STO', 'EA', 0, 0, "Followed by two expressions, assigns the value of the first expression to the second. It is common to have a register as the second expression, but any assignable expression will do."], # 1
    ['CREate', 'S', 'CRE', '#VC(<)C(>)', 0, 0, "Followed by a constant (a grammar) and an expression (a symbol), another constant (a left count), I<left count> rules, another constant (a right count) and I<right count> chunks of code, executes a CREATE statement. This is documented elsewhere."], # 2
    ['DEStroy', 'S', 'DES', '#VC(<)', 0, 0, "Followed by a constant (a grammar) and an expression (a symbol), another constant (a left count), and I<left count> rules, executes a DESTROY statement."], # 3
    ['Make SPlat', 'S', 'MSP', 'EC(V)', 0, 0, "Followed by an expression, the splat code, a number, the count, and I<count> more expressions, generates a splat. The first expression is the splat code, the rest are used to generate the splat message. The splat code determines the correct number of arguments, if the wrong number is provided the message may look weird."], # 4
    ['Double-oh-Seven eXecution', 'S', 'DSX', 'E', 0, 0, "Followed by an expression (which should have value between 0 and 100), it executes the statement with a probability indicated by the expression, between 0% (never) and 100% (always)."], # 5
    ['NOT', 'S', 'NOT', '', 0, 0, "Signals that this statement is initially abstained from. A statement might be abstained from without containing a I<NOT>, or might contain one and not be abstained from, depending on the I<ABL>, I<ABG>, I<REL> and I<REG> executed since the start of the program."], # 6
    ['NeXT', 'S', 'NXT', 'E', 0, 0, "Followed by an expression, a label, stashes the address of the next statement and continues execution at that label. It is an error if the label is multiply defined or not defined at all."], # 7
    ['RESume', 'S', 'RES', 'E', 0, 0, "Followed by an expression, pops that many levels from the stash containing the return addresses for I<NXT>, I<NXL> and I<NXG>. All the addresses except one are then discarded, and execution continues at the last address extracted from the stash."], # 8
    ['FORget', 'S', 'FOR', 'E', 0, 0, "Followed by an expression, pops that many levels from the stash containing the return addresses for I<NXT>, I<NXL> and I<NXG> and throws these addresses in the bit bucket."], # 9
    ['STAsh', 'S', 'STA', 'C(R)', 0, 0, "Followed by a constant (the count) and I<count> registers, STASHes these registers."], # 10
    ['RETrieve', 'S', 'RET', 'C(R)', 0, 0, "Followed by a constant (the count) and I<count> registers, RETRIEVEs these registers."], # 11
    ['IGNore', 'S', 'IGN', 'C(R)', 0, 0, "Followed by a constant (a count) and I<count> registers, ignores the registers."], # 12
    ['REMember', 'S', 'REM', 'C(R)', 0, 0, "Followed by a constant (a count) and I<count> registers, remembers the registers."], # 13
    ['ABstain from Label', 'S', 'ABL', 'E', 0, 0, "Followed by an expression, representing a label, ABSTAINs FROM the corresponding statement(s)."], # 14
    ['ABstain from Gerund', 'S', 'ABG', 'C(O)', 0, 0, "Followed by a constant (the count) and I<count> gerunds, ABSTAINs FROM the corresponding statement(s)."], # 15
    ['REinstate Label', 'S', 'REL', 'E', 0, 0, "Followed by an expression, representing a label, REINSTATEs the corresponding statement(s)."], # 16
    ['REinstate Gerund', 'S', 'REG', 'C(O)', 0, 0, "Followed by a constant (the count) and I<count> gerunds, REINSTATEs the corresponding statement(s)."], # 17
    ['Give UP', 'S', 'GUP', '', 0, 0, "Causes program termination. When used in a compiler module, causes module processing to stop, the compiler will then load the next module or, if no more modules are to be loaded, it will start compiling the program."], # 18
    ['Write IN', 'S', 'WIN', 'C(A)', 0, 0, "Followed by a constant (a count) and I<count> assignable expressions, writes them in."], # 19
    ['Read OUt', 'S', 'ROU', 'C(E)', 0, 0, "Followed by a constant (a count) and I<count> expressions, reads them out."], # 20
    ['LABel', 'S', 'LAB', 'E', 0, 0, "Followed by an expression, indicates this statement's label. If the expression is nonzero, after this statement the ICBM will go looking for corresponding I<CFL> and I<NXL> statements (COME FROMs and NEXT FROMs). It is also used to abstain/reinstate by label."], # 21
    ['Come From Label', 'S', 'CFL', 'E', 0, 0, "Followed by an expression, executes a COME FROM label. The special register I<%CF> determines, amongst other things, whether it is admissible to have multiple COME FROM (and NEXT FROM) all pointing at the same label. The default is to cause a splat; linking with the object I<thick.io> changes this to allow multiple COME FROMs and NEXT FROMs to create threads. See I<CFG> for the other function of the I<%CF> register."], # 22
    ['Come From Gerund', 'S', 'CFG', 'C(O)', 0, 0, "Followed by a constant (the count) and I<count> gerunds (opcodes), executes a COME FROM gerund. The special register I<%CF> determines, amongst other things, whether these statements are really executed or not. The default is not, and linking a program with the object I<come-from-gerund.io> will change this register to allow these statements: this object is normally linked automatically when the program source has a suffix I<.gi>. See I<CFL> for the other function of the I<%CF> register."], # 23
    ['QUAntum statement', 'S', 'QUA', '', 0, 0, "Executes the rest of the statement in \"quantum bit creation\" mode. This means that anything which modifies data will end up creating quantum bits."], # 24
    ['loop: Condition While Body', 'S', 'CWB', 'SS', 0, 0, "Followed by two statements, executes a loop. This implements the (default) loop where the condition is before the WHILE and the body after. The first statement is the body and the second is the condition, not the other way around as one would expect."], # 25
    ['loop: Body While Condition', 'S', 'BWC', 'SS', 0, 0, "Followed by two statements, executes a loop. This implements the (non default) loop where the body is before the WHILE and the condition after. The first statement is the condition and the second is the body, not the other way around as one would expect."], # 26
    ['MaKe Belong', 'S', 'MKB', 'RR', 0, 0, "Followed by two registers, makes the first one belong to the second one."], # 27
    ['No Longer Belong', 'S', 'NLB', 'RR', 0, 0, "Followed by two registers, where the first one must belong to the second one, removes such BELONG TO relationship."], # 28
    ['STUdy', 'S', 'STU', 'EER', 0, 0, "Followed by an expression (the subject), a label (the lecture) and a register (the class), executes a STUDY statement."], # 29
    ['ENRol', 'S', 'ENR', 'C(E)R', 0, 0, "Followed by a count of subjects, I<count> expressions representing the subjects, and a register wishing to study these subjects, looks for a class teaching the subjects and enrols the register there."], # 30
    ['LEArns', 'S', 'LEA', 'ER', 0, 0, "Followed by an expression (the subject), and a register (the student) looks for a lecture where that subject is taught in one of the classes the student is enrolled in. The class register is made to belong to the student until the end of the lecture; execution continues at the label which represents the start of the lecture."], # 31
    ['FINish lecture', 'S', 'FIN', '', 0, 0, "Execution continues after the I<LEA> which took us to the lecture. Also removes the belonging relation of the class to the student."], # 32
    ['GRAduate', 'S', 'GRA', 'R', 0, 0, "Followed by a register (a student) it causes the student to graduate, that is to drop all classes."], # 33
    ['Next From Label', 'S', 'NXL', 'E', 0, 0, "Similar to I<CFL>, but executes a NEXT FROM instead: the difference is that I<NXL> stashes the return address in the same way as I<NXT> does. See I<CFL> and I<NXT>."], # 34
    ['Next From Gerund', 'S', 'NXG', 'C(O)', 0, 0, "Similar to I<CFG>, but executes a NEXT FROM instead: the difference is that I<NXG> stashes the return address in the same way as I<NXT> does. See I<CFG> and I<NXT>."], # 35
    ['CONvert', 'S', 'CON', 'OO', 0, 0, "Followed by two opcodes, converts the first into the second. The two opcodes must be compatible, in the sense that they take the same operands."], # 36
    ['SWAp', 'S', 'SWA', 'OO', 0, 0, "Followed by two opcodes, swaps them. The two opcodes must be compatible, in the sense that they take the same operands."], # 37
    ['compiler BUG', 'S', 'BUG', '#', 0, 0, "This opcode is automatically inserted by the compiler where appropriate. Takes one argument, the bug type (#0 - explainable, #1 - unexplainable)."], # 38
    ['Install DIVersion', 'S', 'DIV', 'EEEE', 0, 0, "Followed by four expressions, representing labels, installs a diversion. The first two labels represent the code to be closed off to execution, and the other two labels represent the actual diversion to be taken."], # 39
    ['Event: Body while Condition', 'S', 'EBC', 'ES', 0, 0, "Followed by an expression and a statement, schedules an event. This implements the (non default) event where the body is before the WHILE and the condition after, and therefore produces a runtime error unless its implementation is CONVERTed to or SWAPped with I<ECB>."], # 40
    ['Event: Condition while Body', 'S', 'ECB', 'ES', 0, 0, "Followed by an expression and a statement, schedules an event. This implements the (default) event where the condition is before the WHILE and the body after."], # 41
    ['FReeZe', 'S', 'FRZ', '', 0, 0, "Freezes the current program by removing the source code and replacing the grammar used to compile it with the \"secondary\" grammar; this means that subsequent I<CRE> and I<DES> will be an error if they cause recompilation. A compiler works by creating a secondary grammar, then freezing itself and then continuing with the user's program, which is compiled using the new grammar just created, rather than the one used to compile the compiler itself."], # 42
    ['SYStem call', 'S', 'SYS', 'EC(S)', 0, 0, "Followed by an expression (the system call number), a count, and I<count> statements (the system call implementation), defines a system call."], # 43
    ['UNdocumented Statement', 'S', 'UNS', '#EEC(E)', 0, 0, "This opcode is documented in the CLC-INTERCAL reference manual."], # 44
    undef, # 45
    undef, # 46
    undef, # 47
    ['TRickle Down', 'S', 'TRD', 'REC(R)', 0, 0, "Set up two registers for trickling down of values"], # 48
    ['TRUss up', 'S', 'TRU', 'C(R)', 0, 0, "Tie values to stop them trickling down"], # 49
    undef, # 50
    undef, # 51
    undef, # 52
    undef, # 53
    undef, # 54
    undef, # 55
    undef, # 56
    undef, # 57
    undef, # 58
    undef, # 59
    ['UNdocumented Assignment', 'S', 'UNA', '#EEC(E)C(A)', 0, 0, "This opcode is documented in the CLC-INTERCAL reference manual."], # 60
    ['REOpen', 'S', 'REO', 'EE', 0, 0, "Followed by two expressions, representing labels, removes the road closure identified by these labels."], # 61
    ['set statement BIT', 'S', 'BIT', 'O', 0, 0, "This opcode should never be executed, and will cause a runtime error; compilers can generate this opcode to update the bitmap of flags which describes a statement. It takes a small constant, between 0 and 31, indicating a bit to set. This statement is executed at the time the object is created and removed from the instruction stream."], # 62
    ['set object FLAg', 'S', 'FLA', '', 0, 0, "This opcode should never be executed, and will cause a runtime error; compilers can generate this opcode to set an object flag, but the opcode will be executed at compile time and removed from the object."], # 63
    ['SPOt', 'R', 'SPO', 'E', 0, 1, "Spot register (e.g. I<.4>)"], # 64
    ['Two SPot', 'R', 'TSP', 'E', 0, 1, "Two spot register (e.g. I<:7>)"], # 65
    ['TAIl', 'R', 'TAI', 'E', 0, 1, "Tail register (e.g. I<,2>). This represents the whole array. See I<SUB> for subscripting."], # 66
    ['HYBrid', 'R', 'HYB', 'E', 0, 1, "Hybrid register (e.g. I<;9>). This represents the whole array. See I<SUB> for subscripting."], # 67
    ['WHirlPool', 'R', 'WHP', 'E', 0, 1, "Whirlpool (e.g. I<\@9>). This represents CLC-INTERCAL's class registers. When used for I/O, it represents the filehandle associated with the class."], # 68
    ['Double-Oh-Seven', 'R', 'DOS', 'E', 0, 1, "Double-oh-seven: a special internal spot register used by the compilers."], # 69
    ['SHark Fin', 'R', 'SHF', 'E', 0, 1, "Shark fin: a special internal tail register used by the compilers (e.g. I<^1>, the arguments given to the program on startup)"], # 70
    ['Crawling HOrror', 'R', 'CHO', 'E', 0, 1, "Crawling horror: a special register holding a compiler, grammar or something similar. Currently, these registers cannot be used directly, they are implicitely used by I<CRE> and I<DES>."], # 71
    undef, # 72
    undef, # 73
    undef, # 74
    undef, # 75
    undef, # 76
    undef, # 77
    undef, # 78
    ['TYPe', 'R', 'TYP', 'RE', 0, 1, "Followed by any register, returns its type. For example, I<TYP> I<SPO> I<136> is equivalent to I<SPO>. It can be useful to find out the type of an indirect register, and is used to translate CLC-INTERCAL's intersection-worm."], # 79
    ['OVerload Register', 'P', 'OVR', 'ER', 0, 1, "Followed by an expression and a register, overloads the register and returns the register itself. Any type of register can be overloaded whether this is a good idea or not."], # 80
    undef, # 81
    ['BeLong', 'P', 'BLM', 'ER', 0, 1, "Followed by a constant N and a register, finds the N-th register it belongs to, and uses that."], # 82
    ['SUBscript', 'P', 'SUB', 'ER', 0, 1, "Followed by an expression and a subscriptable register, it accesses the given subscript. For multidimensional arrays, repeat as in I<SUB> I<131> I<SUB> I<132> I<TAI> I<133> for I<:5 SUB #4 SUB #3>. In addition to hybrid, tail and shark fin registers, whirlpools also accept subscripts, allowing to access the subjects directly."], # 83
    undef, # 84
    undef, # 85
    undef, # 86
    undef, # 87
    undef, # 88
    undef, # 89
    undef, # 90
    undef, # 91
    undef, # 92
    undef, # 93
    undef, # 94
    undef, # 95
    ['MULtiple number', 'E', 'MUL', 'C(E)', 0, 0, "Followed by an expression (the count) and then a number of expressions, represents a ``multiple number''. This can be used to assign to an array, to dimension it (e.g. translating the statement C<;42 <- #3 BY #5> results in the code I<STO> I<MUL> #2 #3 #5 I<HYB> #42). Not assignable."], # 96
    ['STRing', 'E', 'STR', 'C(N)', 0, 0, "Similar to I<MUL>, but used in the special case where all the arguments are constant characters. This may result in internal optimisations and the like. Otherwise it is just a more compact way of using I<MUL> where all arguments are constants and fit in a byte. Not assignable."], # 97
    ['unary BUT', 'E', 'BUT', '#E', 0, 1, "Followed by two expressions, computes the unary BUT of the second expression, preferring the value of the first - so this can also be used for unary 3BUT etc. The special preference value 7, which is invalid for unary BUT, is used to indicate unary OR. Assignable if the second argument is assignable."], # 98
    ['unary AND', 'E', 'AND', 'E', 0, 1, "Followed by an expression, computes its unary AND. This is the same as \"BUT #0\" followed by the same expression but it's faster to execute. Assignable if the operand is assignable."], # 99
    ['unary Subtract Without Borrow', 'E', 'SWB', 'E', 0, 1, "Followed by one expression, it computes the unary subtract without borrow. In base 2, corresponds to the unary exclusive or. Assignable if the argument is."], # 100
    ['unary Inclusive OR', 'E', 'IOR', 'E', 0, 1, "Followed bu an expression, it computes its unary OR. This is the same as \"BUT #7\" followed by the same expression but it's faster to execute. Assignable if the operand is assignable."], # 101
    ['unary Add Without Carry', 'E', 'AWC', 'E', 0, 1, "Followed by one expression, it computes the unary Add without carry; invalid in base 2. Assignable if the argument is."], # 102
    undef, # 103
    ['SELect', 'E', 'SEL', 'EE', 0, 1, "Followed by two expressions, it selects them. Assignable if the arguments are assignable."], # 104
    ['INTerleave', 'E', 'INT', 'EE', 0, 1, "Followed by two expressions, interleaves them. Assignable if both arguments are assignable."], # 105
    ['NUMber', 'E', 'NUM', 'R', 0, 1, "Followed by a register, returns its number. So for example I<NUM> I<SPO> I<#2> would be the same as I<#2>. This is more useful when the register provided is reached using I<BLM>. Assignable."], # 106
    ['OVerload Many', 'E', 'OVM', 'EE', 0, 1, "Followed by two expressions, overloads a range of registers. Note that all \"user\" types of registers are overloaded, but \"special\" registers, including whirlpool, are not. The range is determined by uninterleaving the second argument. See also I<OVR>. Assignable."], # 107
    undef, # 108
    ['SPLat', 'E', 'SPL', '', 0, 1, "Returns the code of the last splat. This is only useful if the program is quantum or threaded, otherwise it won't be executing after a splat. If there was no splat, generates one. Assignable, but in this case it unconditionally splats for obvious reasons."], # 109
    ['Unary DiVide', 'E', 'UDV', 'E', 0, 1, "The \"most useless\" operation, but surely somebody will find a use for it. This operation can be considered arithmetic or bitwise, depending on the value of special register I<%DM>."], # 110
    ['Reverse SElect', 'E', 'RSE', 'EE', 0, 1, "Like I<SEL>, but swaps its operands."], # 111
    ['Reverse INterleave', 'E', 'RIN', 'EE', 0, 1, "Like I<INT>, but swaps its operands."], # 112
    ['UNdocumented Expression', 'E', 'UNE', 'EEC(E)', 0, 0, "This opcode is documented in the CLC-INTERCAL reference manual."], # 113
    undef, # 114
    undef, # 115
    undef, # 116
    undef, # 117
    undef, # 118
    undef, # 119
    undef, # 120
    undef, # 121
    undef, # 122
    undef, # 123
    undef, # 124
    undef, # 125
    ['Half Spot Number', '#', 'HSN', 'N', 1, 1, "Followed by a second byte, represents the value of that byte."], # 126
    ['One Spot Number', '#', 'OSN', 'NN', 1, 1, "Followed by two bytes, represents the 16 bit number which has the first such byte as higher significant half, and the second byte as lower significant half."], # 127
);

my %bytedecode = (
    ABG => 15,
    ABL => 14,
    AND => 99,
    AWC => 102,
    BIT => 62,
    BLM => 82,
    BUG => 38,
    BUT => 98,
    BWC => 26,
    CFG => 23,
    CFL => 22,
    CHO => 71,
    CON => 36,
    CRE => 2,
    CWB => 25,
    DES => 3,
    DIV => 39,
    DOS => 69,
    DSX => 5,
    EBC => 40,
    ECB => 41,
    ENR => 30,
    FIN => 32,
    FLA => 63,
    FOR => 9,
    FRZ => 42,
    GRA => 33,
    GUP => 18,
    HSN => 126,
    HYB => 67,
    IGN => 12,
    INT => 105,
    IOR => 101,
    LAB => 21,
    LEA => 31,
    MKB => 27,
    MSP => 4,
    MUL => 96,
    NLB => 28,
    NOT => 6,
    NUM => 106,
    NXG => 35,
    NXL => 34,
    NXT => 7,
    OSN => 127,
    OVM => 107,
    OVR => 80,
    QUA => 24,
    REG => 17,
    REL => 16,
    REM => 13,
    REO => 61,
    RES => 8,
    RET => 11,
    RIN => 112,
    ROU => 20,
    RSE => 111,
    SEL => 104,
    SHF => 70,
    SPL => 109,
    SPO => 64,
    STA => 10,
    STO => 1,
    STR => 97,
    STS => 0,
    STU => 29,
    SUB => 83,
    SWA => 37,
    SWB => 100,
    SYS => 43,
    TAI => 66,
    TRD => 48,
    TRU => 49,
    TSP => 65,
    TYP => 79,
    UDV => 110,
    UNA => 60,
    UNE => 113,
    UNS => 44,
    WHP => 68,
    WIN => 19,
);

my @bc_list = qw(
    ABG ABL AND AWC BIT BLM BUG BUT BWC CFG CFL CHO CON CRE CWB DES DIV DOS
    DSX EBC ECB ENR FIN FLA FOR FRZ GRA GUP HSN HYB IGN INT IOR LAB LEA MKB
    MSP MUL NLB NOT NUM NXG NXL NXT OSN OVM OVR QUA REG REL REM REO RES RET
    RIN ROU RSE SEL SHF SPL SPO STA STO STR STS STU SUB SWA SWB SYS TAI TRD
    TRU TSP TYP UDV UNA UNE UNS WHP WIN
);

sub BC_ABG () { 15; }
sub BC_ABL () { 14; }
sub BC_AND () { 99; }
sub BC_AWC () { 102; }
sub BC_BIT () { 62; }
sub BC_BLM () { 82; }
sub BC_BUG () { 38; }
sub BC_BUT () { 98; }
sub BC_BWC () { 26; }
sub BC_CFG () { 23; }
sub BC_CFL () { 22; }
sub BC_CHO () { 71; }
sub BC_CON () { 36; }
sub BC_CRE () { 2; }
sub BC_CWB () { 25; }
sub BC_DES () { 3; }
sub BC_DIV () { 39; }
sub BC_DOS () { 69; }
sub BC_DSX () { 5; }
sub BC_EBC () { 40; }
sub BC_ECB () { 41; }
sub BC_ENR () { 30; }
sub BC_FIN () { 32; }
sub BC_FLA () { 63; }
sub BC_FOR () { 9; }
sub BC_FRZ () { 42; }
sub BC_GRA () { 33; }
sub BC_GUP () { 18; }
sub BC_HSN () { 126; }
sub BC_HYB () { 67; }
sub BC_IGN () { 12; }
sub BC_INT () { 105; }
sub BC_IOR () { 101; }
sub BC_LAB () { 21; }
sub BC_LEA () { 31; }
sub BC_MKB () { 27; }
sub BC_MSP () { 4; }
sub BC_MUL () { 96; }
sub BC_NLB () { 28; }
sub BC_NOT () { 6; }
sub BC_NUM () { 106; }
sub BC_NXG () { 35; }
sub BC_NXL () { 34; }
sub BC_NXT () { 7; }
sub BC_OSN () { 127; }
sub BC_OVM () { 107; }
sub BC_OVR () { 80; }
sub BC_QUA () { 24; }
sub BC_REG () { 17; }
sub BC_REL () { 16; }
sub BC_REM () { 13; }
sub BC_REO () { 61; }
sub BC_RES () { 8; }
sub BC_RET () { 11; }
sub BC_RIN () { 112; }
sub BC_ROU () { 20; }
sub BC_RSE () { 111; }
sub BC_SEL () { 104; }
sub BC_SHF () { 70; }
sub BC_SPL () { 109; }
sub BC_SPO () { 64; }
sub BC_STA () { 10; }
sub BC_STO () { 1; }
sub BC_STR () { 97; }
sub BC_STS () { 0; }
sub BC_STU () { 29; }
sub BC_SUB () { 83; }
sub BC_SWA () { 37; }
sub BC_SWB () { 100; }
sub BC_SYS () { 43; }
sub BC_TAI () { 66; }
sub BC_TRD () { 48; }
sub BC_TRU () { 49; }
sub BC_TSP () { 65; }
sub BC_TYP () { 79; }
sub BC_UDV () { 110; }
sub BC_UNA () { 60; }
sub BC_UNE () { 113; }
sub BC_UNS () { 44; }
sub BC_WHP () { 68; }
sub BC_WIN () { 19; }

# line 62

sub add_bytecode {
    @_ == 7 or croak
	"Usage: add_bytecode(SOURCE, NAME, DESCR, TYPE, NUMBER, ARGS, DOC)";
    my ($source, $name, $descr, $type, $number, $args, $doc) = @_;
    my ($const, $assign);
    $type = uc($type);
    if ($type eq 'S' || $type eq 'E') {
	$const = $assign = 0;
    } elsif ($type eq 'A') {
	$const = 0;
	$assign = 1;
	$type = 'E';
    } elsif ($type eq 'R' || $type eq 'P') {
	$const = 0;
	$assign = 1;
    } elsif ($type eq 'C') {
	$const = 1;
	$assign = 1;
	$type = '#';
    } else {
	croak "Invalid TYPE: $type";
    }
    $name = uc($name);
    $number += 0;
    $number < 0 || $number >= NUM_OPCODES
	and croak "Invalid bytecode number: $number";
    $bytecodes[$number]
	and croak "Duplicate bytecode number: $number";
    exists $bytedecode{$name}
	and croak "Duplicate bytecode name: $name";
    $bytedecode{$name} = $number;
    $bytecodes[$number] = [$descr, $type, $name, $args, $const, $assign, $doc, $source];
    push @bc_list, $name;
    push @EXPORT_OK, "BC_$name";
    push @{$EXPORT_TAGS{BC}}, "BC_$name";
    no strict;
    *{"BC_$name"} = sub { $number };
}

sub bc_list () {
    @bc_list;
}

sub BC ($) {
    my ($val) = @_;
    croak "Invalid undefined value" unless defined $val;
    my $orig = $val;
    $val < BYTE_SHIFT
	and return ($val + NUM_OPCODES);
    $val < OPCODE_RANGE
	and return (BC_HSN, $val);
    my $div = int($val / OPCODE_RANGE);
    $div < OPCODE_RANGE
	and return (BC_OSN, $div, $val % OPCODE_RANGE);
    croak "Invalid value $orig: does not fit in one spot";
}

sub bytecode ($) {
    my ($name) = @_;
    $name =~ /^\d+$/ && $name < BYTE_SHIFT ? ($name + NUM_OPCODES)
					   : $bytedecode{$name};
}

# convert bytecode to name, or in list context to:
# (name, description, type, opcode, pattern, is_constant?, is_assignable?)
#     0            1     2       3        4             5               6
sub bytedecode ($) {
    my ($b) = @_;
    if ($b >= NUM_OPCODES) {
	my $n = $b - NUM_OPCODES;
	return () if $n >= BYTE_SHIFT;
	return "#$n" unless wantarray;
	return ("#$n", 'Constant', '#', $b, '', 1, 1);
    } else {
	return () unless $bytecodes[$b];
	my $p = $bytecodes[$b];
	return $p->[2] unless wantarray;
	return ($p->[2], $p->[0], $p->[1], $b, $p->[3], $p->[4], $p->[5]);
    }
}

sub bytename ($) {
    my ($b) = @_;
    if ($b >= NUM_OPCODES) {
	my $n = $b - NUM_OPCODES;
	return () if $n >= BYTE_SHIFT;
	return "#$n";
    } else {
	return () unless $bytecodes[$b];
	my $p = $bytecodes[$b];
	return $p->[2];
    }
}

sub BCget ($$$) {
    # don't make a copy of $_[0], it's likely to be a long string and we only want
    # at most 3 bytes from it
    my ($cp, $ep) = @_[1, 2];
    $$cp >= $ep and faint(SP_INTERNAL, 'BCget called after end of code');
    my $byte = vec($_[0], $$cp++, 8);
    if ($byte >= NUM_OPCODES) {
	return $byte - NUM_OPCODES;
    }
    if ($byte == BC_HSN) {
	$$cp >= $ep and faint(SP_INTERNAL, 'BCget: missing constant after HSN');
	return vec($_[0], $$cp++, 8);
    }
    if ($byte == BC_OSN) {
	$$cp + 1 >= $ep and faint(SP_INTERNAL, 'BCget: missing constant after OSN');
	my $nx = vec($_[0], $$cp++, 8) << 8;
	return $nx | vec($_[0], $$cp++, 8);
    }
    faint(SP_INTERNAL, sprintf("BCget: unknown opcode 0x%02x", $byte));
}

sub is_constant ($) {
    my ($byte) = @_;
    return 1 if $byte >= NUM_OPCODES ||
		$byte == BC_HSN ||
		$byte == BC_OSN;
    return 0;
}

sub bc_skip ($$$) {
    my (undef, $start, $end) = @_; # $_[0] is code but we don't want to copy it
    $$start >= $end and return undef;
    my $byte = vec($_[0], $$start++, 8);
    $byte >= NUM_OPCODES and return 1;
    $bytecodes[$byte] or return undef;
    my $expect = $bytecodes[$byte][3];
    while ($expect ne '') {
	my $e = substr($expect, 0, 1, '');
	if ($e eq '#' || $e eq 'C') {
	    # constant, optionally followed by (submatch)
	    $$start >= $end and return undef;
	    my $byte = vec($_[0], $$start++, 8);
	    if ($expect ne '' && substr($expect, 0, 1) eq '(') {
		my $level = 1;
		my $pos = 1;
		while ($level > 0) {
		    $pos >= length $expect and return undef;
		    my $c = substr($expect, $pos++, 1);
		    if ($c eq '(') {
			$level++;
		    } elsif ($c eq ')') {
			$level--;
		    }
		}
		if ($byte >= NUM_OPCODES) {
		    $byte -= NUM_OPCODES;
		} elsif ($byte == BC_HSN) {
		    $$start >= $end and return undef;
		    $byte = vec($_[0], $$start++, 8);
		} elsif ($byte == BC_OSN) {
		    $$start + 1 >= $end and return undef;
		    # can't use vec(..., 16) because it may not be 16-bit aligned
		    $byte = vec($_[0], $$start++, 8) << 8;
		    $byte |= vec($_[0], $$start++, 8);
		} else {
		    return undef;
		}
		my $subarg = substr($expect, 0, $pos, '');
		$byte and $expect = (substr($subarg, 1, $pos - 2) x $byte) . $expect;
		next;
	    }
	    $byte >= NUM_OPCODES and next;
	    if ($byte == BC_HSN) {
		$$start >= $end and return undef;
		$$start++;
		next;
	    }
	    if ($byte == BC_OSN) {
		$$start += 2;
		$$start > $end and return undef;
		next;
	    }
	    return undef;
	}
	if ($e eq 'S' || $e eq 'E' || $e eq 'R' || $e eq 'P' || $e eq 'A' || $e eq 'V') {
	    # statment, expression, register, asignable or symbol: they
	    # all match a generic bytecode sequence
	    $$start >= $end and return undef;
	    my $byte = vec($_[0], $$start++, 8);
	    $byte >= NUM_OPCODES and next;
	    $bytecodes[$byte] or return undef;
	    $expect = $bytecodes[$byte][3] . $expect;
	    next;
	}
	if ($e eq 'O') {
	    # gerund or similar symbol: anything other than HSN or OSN means
	    # a 1-byte constant
	    $$start >= $end and return undef;
	    my $byte = vec($_[0], $$start++, 8);
	    if ($byte == BC_HSN) {
		$$start >= $end and return undef;
		$$start++;
		next;
	    }
	    if ($byte == BC_OSN) {
		$$start += 2;
		$$start > $end and return undef;
		next;
	    }
	    next;
	}
	if ($e eq 'N') {
	    # any byte
	    $$start >= $end and return undef;
	    $$start++;
	    next;
	}
	if ($e eq '<') {
	    # left grammar rule: count, position, symbol or string
	    $expect = '##E' . $expect;
	    next;
	}
	if ($e eq '>') {
	    # right grammar rule: a constant follows determining what's next
	    $$start >= $end and return undef;
	    my $byte = vec($_[0], $$start++, 8);
	    if ($byte == NUM_OPCODES || $byte == NUM_OPCODES + 1 || $byte == NUM_OPCODES + 3 || $byte == NUM_OPCODES + 6) {
		# position, symbol / number
		$expect = 'EE' . $expect;
		next;
	    }
	    if ($byte == NUM_OPCODES + 4) {
		# length; block of bytecode
		$$start >= $end and return undef;
		$byte = vec($_[0], $$start++, 8);
		if ($byte >= NUM_OPCODES) {
		    $byte -= NUM_OPCODES;
		} elsif ($byte == BC_HSN) {
		    $$start >= $end and return undef;
		    $byte = vec($_[0], $$start++, 8);
		} elsif ($byte == BC_OSN) {
		    $$start + 1 >= $end and return undef;
		    # can't use vec(..., 16) because it may not be 16-bit aligned
		    $byte = vec($_[0], $$start++, 8) << 8;
		    $byte |= vec($_[0], $$start++, 8);
		} else {
		    return undef;
		}
		$$start += $byte;
		$$start > $end and return undef;
		next;
	    }
	    if ($byte == NUM_OPCODES + 15) {
		# "splat", no other data
		next;
	    }
	    # unknown type
	    return undef;
	}
    }
    return 1;
}

sub bc_forall {
    @_ == 5
	or croak "Usage: bc_forall(PATTERN, CODE, START, END, CLOSURE)";
    my ($pattern, undef, $start, $end, $closure) = @_;
    $start ||= 0;
    $end = length($_[1]) if not defined $end;
    return undef if $start >= $end || $start < 0;
    my $np = '';
    while ($pattern =~ s/^(.*?)C\(/(/) {
	my $a = $1;
	$a =~ s/(.)/$1\x01/g;
	$np .= $a . 'C';
	$np .= '(' . _args('forall', \$pattern) . ')';
	$np .= "\01";
    }
    $pattern =~ s/(.)/$1\x01/g;
    $pattern = "\x01" if $pattern eq '';
    $np .= $pattern;
    _forall($np, $_[1], $start, $end, $closure);
}

my %typemap = (
    'S' => { 'S' => 0 },
    'O' => { 'S' => 0 },
    'E' => { 'E' => 0, 'R' => 0, 'P' => 0, '#' => 0 },
    'A' => { 'E' => 0, 'R' => 0, 'P' => 0, '#' => 0 },
    'R' => { 'R' => 0, 'P' => 0 },
    'P' => { 'R' => 0, 'P' => 0 },
    'V' => { 'R' => 0, 'P' => 0, 'V' => 0 },
    '#' => { '#' => 0 },
    'C' => { '#' => 0 },
    'Z' => { 'S' => 0, 'E' => 0, 'R' => 0, 'P' => 0, '#' => 0 },
    '*' => { 'S' => 0, 'E' => 0, 'R' => 0, 'P' => 0, '#' => 0 },
);

sub _args {
    my ($name, $pattern) = @_;
    faint(SP_BCMATCH, $name, 'Missing (') if $$pattern !~ s/^\(//;
    my $count = 1;
    my $result = '';
    while ($count > 0) {
	$$pattern =~ s/^([^\(\)]*)([\(\)])//
	    or faint(SP_BCMATCH, $name, 'Missing )');
	$count++ if $2 eq '(';
	$count-- if $2 eq ')';
	$result .= $1 . ($count ? $2 : '');
    }
    $result;
}

sub _forall {
    my ($pattern, undef, $sc, $ep, $closure) = @_;
    my $osc = $sc;
    MATCH: while ($pattern ne '') {
	my $e = substr($pattern, 0, 1, '');
	if ($e eq "\x00") {
	    $closure->(undef, '>') if $closure;
	    next MATCH;
	}
	if ($e eq "\x01") {
	    $closure->($sc, undef) if $closure;
	    next MATCH;
	}
	faint(SP_INTERNAL, '_forall: reading past end of code') if $sc >= $ep;
	my $v = vec($_[1], $sc, 8);
	if (exists $typemap{$e}) {
	    # check next opcode is correct type
	    my ($op, $type, $args, $const);
	    if ($v >= NUM_OPCODES && $e ne 'O') {
		$op = '#' . ($v - NUM_OPCODES);
		$type = '#';
		$args = '';
		$const = 1;
	    } else {
		$v %= NUM_OPCODES; # so gerunds can be small constants as well as opcodes
		$bytecodes[$v] or faint(SP_INTERNAL, "_forall: $e: invalid gerund $v");
		my $p = $bytecodes[$v];
		$op = $p->[2];
		$type = $p->[1];
		$args = $p->[3];
		$const = $p->[4];
	    }
	    faint(SP_INTERNAL, "_forall: $e: unknown type $type")
		unless exists $typemap{$e}{$type} ||
		       (($v == BC_MUL || $v == BC_STR) && exists $typemap{$e}{V});
	    if ($e eq 'O' && $const) {
		# inlining a stripped-down version of BCget because this is like
		# the inner loop of an inner loop
		if ($v < NUM_OPCODES) {
		    if ($v == BC_HSN) {
			$sc < $ep or return ();
			$sc++;
		    } elsif ($v == BC_OSN) {
			$sc += 2;
			$sc <= $ep or return ();
		    } else {
			return ();
		    }
		}
	    } elsif ($type eq '#' && $e ne '*') {
		my $num = BCget($_[1], \$sc, $ep);
		$closure->($v, "#$num") if $closure;
		if ($e eq 'C') {
		    $args = _args('count', \$pattern) x $num;
		    $args .= "\x00";
		    $closure->(undef, '<') if $closure;
		} else {
		    $args = '';
		}
	    } else {
		$sc++;
		$args = '' if $e eq 'O' || $e eq '*';
		$closure->($v, $op) if $closure;
	    }
	    $pattern = $args . $pattern;
	    next MATCH;
	} elsif ($e eq 'N') {
	    # any number
	    $closure->($v, "N$v") if $closure;
	    $sc++;
	} elsif ($e eq '<') {
	    # left grammar element
	    my $count = BCget($_[1], \$sc, $ep);
	    my $num = BCget($_[1], \$sc, $ep);
	    if ($num == 0) {
		$closure->(undef, '?<') if $closure;
	    } elsif ($num == 1 || $num == 2) {
		$closure->(undef, ',<') if $closure;
	    } else {
		$closure->(undef, ',!<') if $closure;
	    }
	    if ($count && $closure) {
		$closure->(undef, $count == 65535 ? '*' : $count);
	    }
	    $pattern = "E\x00" . $pattern;
	    next MATCH;
	} elsif ($e eq '>') {
	    # right grammar element
	    my $num = BCget($_[1], \$sc, $ep);
	    if ($num == 0 || $num == 6) {
		my $count = BCget($_[1], \$sc, $ep);
		if ($count && $closure) {
		    $closure->(undef, $count);
		}
		$closure->($v, $num ? '!<' : '?<') if $closure;
		$pattern = "E\x00" . $pattern;
		next MATCH;
	    }
	    if ($num == 1 || $num == 2) {
		$closure->($v, ',<') if $closure;
		my $count = BCget($_[1], \$sc, $ep);
		if ($count && $closure) {
		    $closure->(undef, $count);
		}
		$pattern = "E\x00" . $pattern;
		next MATCH;
	    }
	    if ($num == 3 || $num == 7) {
		$closure->($v, ',!<') if $closure;
		my $count = BCget($_[1], \$sc, $ep);
		if ($count && $closure) {
		    $closure->(undef, $count);
		}
		$pattern = "E\x00" . $pattern;
		next MATCH;
	    }
	    if ($num == 4) {
		$num = BCget($_[1], \$sc, $ep);
		my $se = $sc + $num;
		$se <= $ep
		    or faint(SP_INTERNAL, '_forall: end of code reached in nested call');
		if ($closure) {
		    $closure->(undef, '=<');
		    while ($sc < $se) {
			$sc += _forall('*', $_[1], $sc, $se, $closure);
		    }
		    $closure->(undef, '>');
		} else {
		    $sc = $se;
		}
		next MATCH;
	    }
	    if ($num == 15) {
		$closure->($v, '*') if $closure;
		next MATCH;
	    }
	    faint(SP_INTERNAL, $num, "_forall: invalid nested call parameter $num");
	} elsif ($e eq '[') {
	    # XXX left optimise element
	    faint(SP_TODO, 'match on [');
	} elsif ($e eq ']') {
	    # XXX right optimise element
	    faint(SP_TODO, 'match on ]');
	} else {
	    faint(SP_BCMATCH, 'type', $e);
	}
    }
    $sc - $osc;
}

sub all_opcodes {
    grep { defined } @bytecodes;
}

1;

__END__

=pod

=head1 NAME

Language::INTERCAL::Bytecode - intermediate language

=head1 DESCRIPTION

The CLC-INTERCAL compiler works by producing bytecode from the
program source; this bytecode can be interpreted to execute the
program immediately; alternatively, a backend can produce something
else from the bytecode, for example C or Perl source code which can
then be compiled to your computer's native object format.

The compiler itself is just some more bytecode. Thus, to produce the
compiler you need a compiler compiler, and to produce that you need
a compiler compiler compiler; to produce the latter you would need
a compiler compiler compiler compiler, and so on to infinity. To
simplify the programmer's life (eh?), the compiler compiler is able
to compile itself, and is therefore identical to the compiler compiler
compiler (etcetera).

The programmer can start the process because a pre-compiled compiler
compiler, in the form of bytecode, is generated somehow while installing
CLC-INTERCAL: this compiler compiler then is able to compile all
other compilers, as well as to rebuild itself if need be.

See the online manual or the HTML documentation included with the
distribution for more information about this.

=head1 SEE ALSO

A qualified psychiatrist

=head1 AUTHOR

Claudio Calvelli - compiler (whirlpool) intercal.org.uk
(Please include the word INTERLEAVING in the subject when emailing that
address, or the email may be ignored)

